/*
 *  This file is part of M-Index library: http://disa.fi.muni.cz/m-index/
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.algorithms;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import messif.algorithms.Algorithm;
import messif.algorithms.AlgorithmMethodException;
import messif.algorithms.NavigationProcessor;
import messif.algorithms.impl.AbstractNavigationProcessor;
import messif.algorithms.impl.MultipleOverlaysAlgorithm;
import messif.objects.LocalAbstractObject;
import messif.operations.AbstractOperation;
import messif.operations.Approximate;
import messif.operations.RankingQueryOperation;
import messif.operations.data.BulkInsertOperation;
import messif.operations.data.InsertOperation;
import messif.operations.query.KNNQueryOperation;
import messif.operations.query.RangeQueryOperation;
import mindex.MIndexMultiplePP;
import mindex.MIndexPPCalculator;

/**
 * Indexing algorithm that processes operation by passing them to encapsulated collection
 * of M-Index algorithms.
 * 
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public class MultipleMIndexAlgorithm extends MultipleOverlaysAlgorithm {
    
    /** Class id for serialization. */
    private static final long serialVersionUID = 201001L;

    /** Internal list of algorithms (the same as encapsulated ones from the constructor). */
    private transient MIndexAlgorithm[] algorithms;
    
    /** A simple lock for periodic algorithm serialization. */
    protected transient Lock serializingLock = new ReentrantLock();
    

    /**
     * Creates a new multi-algorithm overlay for the given array of algorithms.
     * @param algorithms the algorithms on which the operations are processed
     * @param cloneAsynchronousOperation the flag whether to clone the operation for asynchronous processing
     * @param processInThreads only if this flag is true, the operations are processed asynchronously on each encapsulated algorithm
     */
    public MultipleMIndexAlgorithm(Collection<? extends Algorithm> algorithms, boolean cloneAsynchronousOperation) {
        super(algorithms, cloneAsynchronousOperation);        
        setAlgorithms(algorithms);
    }

    /**
     * Creates a new multi-algorithm overlay for the given collection of algorithms.
     * @param algorithms the algorithms on which the operations are processed
     * @param cloneAsynchronousOperation the flag whether to clone the operation for asynchronous processing
     * @param processInThreads only if this flag is true, the operations are processed asynchronously on each encapsulated algorithm
     */
    @Algorithm.AlgorithmConstructor(description = "Constructor with created algorithms", arguments = {"array of running algorithms", "t/f if operation should be clonned before running"})
    public MultipleMIndexAlgorithm(Algorithm[] algorithms, boolean cloneAsynchronousOperation) {
        this(Arrays.asList(algorithms), cloneAsynchronousOperation);
    }

    /**
     * Initializes the internal array of encapsulated algorithms with the given collection.
     * @param algorithms the algorithms for the array
     */
    private void setAlgorithms(Collection<? extends Algorithm> algorithms) {
        this.algorithms = algorithms.toArray(new MIndexAlgorithm[algorithms.size()]);
    }

    /** Deserialize and set the transient fields. */
    private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
        setAlgorithms(getAlgorithms());
        serializingLock = new ReentrantLock();        
    }
    
    @Override
    public <E extends AbstractOperation> List<Class<? extends E>> getSupportedOperations(Class<? extends E> subclassToSearch) {
        if (algorithms.length > 0) {
            return algorithms[0].getSupportedOperations(subclassToSearch);
        }
        return super.getSupportedOperations(subclassToSearch);
    }
    
    /**
     * Checks all the M-Index algorithms, if their dynamic cell tree was not modified and
     *  serializes the whole algorithm if it was changed. THis method is typically called
     *  periodically.
     * @param serializationFile file to serialize the whole algorithm into
     * @throws IOException if the serialization failed
     */
    public void checkModifiedAndStore(String serializationFile) throws IOException {
        serializingLock.lock();
        try {
            boolean anyModified = false;
            for (MIndexAlgorithm mIndexAlgorithm : algorithms) {
                anyModified = anyModified || mIndexAlgorithm.isDynamicTreeModified();
            }
            if (anyModified) {
                // Store algorithm to file
                this.storeToFile(serializationFile);
            }
        } finally {
            serializingLock.unlock();
        }
    }

    @Override
    protected void onStoreToFile(String filepath) {
        super.onStoreToFile(filepath);
        for (MIndexAlgorithm mIndexAlgorithm : algorithms) {
            mIndexAlgorithm.setDynamicTreeModified(false);
        }
    }

    public Collection<MIndexPPCalculator> getAllPPCalculators() {
        ArrayList<MIndexPPCalculator> pPCalculators = new ArrayList<>(algorithms.length);
        for (MIndexAlgorithm alg : algorithms) {
            pPCalculators.add(alg.getPPCalculator());
        }
        return pPCalculators;
    }
    
    /** 
     * Clears (removes) all the data stored in all referenced M-Index algorithms. The structure of the 
     *  dynamic cell trees remains untouched.
     */
    public void clearData() throws AlgorithmMethodException {
        for (MIndexAlgorithm mIndexAlgorithm : algorithms) {
            mIndexAlgorithm.clearData();
        }
    }
    
    @Override
    public NavigationProcessor<? extends AbstractOperation> getNavigationProcessor(AbstractOperation operation) {
        if ((operation instanceof InsertOperation) || (operation instanceof BulkInsertOperation)) {
//            return super.getNavigationProcessor(operation);
            return new AbstractNavigationProcessor<AbstractOperation, Algorithm>(operation, true, true, getAlgorithms()) {
                @Override
                protected AbstractOperation processItem(AbstractOperation operation, Algorithm algorithm) throws AlgorithmMethodException {
                    try {
                        // check and modify the M-Index keys, if already preset
                        int algIndex = Arrays.asList(algorithms).indexOf(algorithm);
                        if (operation instanceof InsertOperation) {
                            LocalAbstractObject insertedObject = ((InsertOperation) operation).getInsertedObject();
                            setRelevantMIndexPP(insertedObject, algIndex);
                        } else if (operation instanceof BulkInsertOperation) {
                            for (LocalAbstractObject obj : ((BulkInsertOperation) operation).getInsertedObjects()) {
                                if (!setRelevantMIndexPP(obj, algIndex)) {
                                    break;
                                }
                            }
                        }
                        return algorithm.executeOperation(operation);
                    } catch (NoSuchMethodException e) {
                        throw new AlgorithmMethodException(e);
                    }
                }

                protected boolean setRelevantMIndexPP(LocalAbstractObject insertedObject, int algIndex) throws IllegalArgumentException, NullPointerException {
                    MIndexMultiplePP multiplePP = insertedObject.getDistanceFilter(MIndexMultiplePP.class);
                    if (multiplePP == null) {
                        return false;
                    }
                    insertedObject.unchainFilter(multiplePP);
                    insertedObject.chainFilter(multiplePP.getPivotPermutation(algIndex), true);
                    return true;
                }
            };
        }
        if (operation instanceof RankingQueryOperation) {
            // Update answer flag in order to prevent duplicates in the merged answer
            ((RankingQueryOperation)operation).setAnswerIgnoringDuplicates(true);
        }            
        if (operation instanceof Approximate || operation instanceof KNNQueryOperation || operation instanceof RangeQueryOperation) {
            if (operation instanceof Approximate) {
                ((Approximate) operation).setLocalSearchParam(((Approximate) operation).getLocalSearchParam() / getAlgorithmsCount());
            }
            return super.getNavigationProcessor(operation);
        }
        // TODO: may be create own processor (handle statistics, if on)
        return null;
    }

}
