/*
 *  This file is part of M-Index library: http://disa.fi.muni.cz/m-index/
 *
 *  M-Index library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  M-Index library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with M-Index library.  If not, see <http://www.gnu.org/licenses/>.
 */
package mindex.processors;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.Callable;
import messif.algorithms.AlgorithmMethodException;
import messif.algorithms.AsynchronousNavigationProcessor;
import messif.objects.LocalAbstractObject;
import messif.operations.AbstractOperation;
import messif.operations.data.BulkInsertOperation;
import messif.operations.data.InsertOperation;
import mindex.MIndexPP;
import mindex.MetricIndex;
import mindex.MetricIndexes;
import mindex.navigation.SplitConfiguration;
import mindex.navigation.VoronoiLeafCell;

/**
 * Implementation of NavigationProcessor interface that can handle both
 * InsertOperation and BulkInsertOperation.
 * @author David Novak, Masaryk University, Brno, Czech Republic, novak.david@gmail.com
 */
public class InsertOperationNavigationProcessor implements AsynchronousNavigationProcessor<AbstractOperation> {

    /** Maps objects that should be inserted into the index to appropriate leaves */
    private final Map<VoronoiLeafCell, List<LocalAbstractObject>> cellObjectMap = new HashMap<>();   
    
    /** How many LeafCells were already processed (all objects that should be inserted into this leaf were inserted) */
    private int processed = 0;

    /** How many LeafCells were retrieved from the {@link #cellObjectMap} (used for counting # of running "operations") */
    private int retrievedFromQueue = 0;
    
    /** Either BulkInsertOperation or simple InsertOperation that will be processed by this processor */
    private AbstractOperation operation;
    
    /** MetricIndex (NavigationDirectory) into which the objects should be inserted */
    private final MetricIndex mIndex;

    /**
     * Inits the processor for a single-object insert operation. The object is initialized
     *  by M-Index (set M-Index pivot permutation) and added to the cell map.
     * @param operation single-object insert operation
     * @param mIndex M-Index object
     * @throws AlgorithmMethodException if the initialization goes wrong
     */
    public InsertOperationNavigationProcessor(InsertOperation operation, MetricIndex mIndex) throws AlgorithmMethodException {
        this((AbstractOperation) operation, mIndex);
        LocalAbstractObject obj = operation.getInsertedObject();
        mIndex.initPP(obj);
        addToCellObjectMap(obj);
    }

    /**
     * Inits the processor for a multi-object bulk-insert operation. The objects are initialized
     *  by M-Index (set M-Index pivot permutation) and added to the cell map.
     * @param operation single-object insert operation
     * @param mIndex M-Index object
     * @throws AlgorithmMethodException if the initialization goes wrong
     */
    public InsertOperationNavigationProcessor(BulkInsertOperation operation, MetricIndex mIndex) throws AlgorithmMethodException {
        this((AbstractOperation) operation, mIndex);
        initializeObjects(operation.getInsertedObjects());
    }

    private InsertOperationNavigationProcessor(AbstractOperation operation, MetricIndex mIndex) {
        this.operation = operation;
        this.mIndex = mIndex;
    }

    /**
     * A process step of this Processor means inserting all LocalAbstractObjects
     * that are mapped to a MIndexLeafCell into its Bucket
     * @return true  if the are unprocessed StructureLeafNodes
     *         false otherwise
     */
    @Override
    public boolean isFinished() {
        return cellObjectMap.isEmpty();
    }

    /**
     * Returns the next processing item in the queue.
     * @return the next processing item in the queue or <tt>null</tt> if the queue is empty
     * @throws InterruptedException if the waiting for the next item in the queue has been interrupted
     */
    protected Entry<VoronoiLeafCell, List<LocalAbstractObject>> getNextProcessingItem() throws InterruptedException {
        synchronized (cellObjectMap) {
            while (cellObjectMap.isEmpty()) {
                if (processed >= retrievedFromQueue) {
                    return null;
                }
                cellObjectMap.wait();
            }
            Entry<VoronoiLeafCell, List<LocalAbstractObject>> objectsToBeInserted = cellObjectMap.entrySet().iterator().next();
            cellObjectMap.remove(objectsToBeInserted.getKey());
            cellObjectMap.notifyAll();
            retrievedFromQueue ++;
            return objectsToBeInserted;
        }
    }
    
    /**
     * Processes next MIndexLeafCell and inserts LocalAbstractObjects that
     * are mapped to it into its Bucket.
     * 
     * @return the number of inserted objects
     */
    @Override
    public boolean processStep() throws AlgorithmMethodException, InterruptedException {
        Entry<VoronoiLeafCell, List<LocalAbstractObject>> objectsToBeInserted = getNextProcessingItem();
        if (objectsToBeInserted == null) {
            return false;
        }
        // Insert the objects into the MIndexLeafCell
        try {
            processObjectsOnLeaf(objectsToBeInserted.getKey(), objectsToBeInserted.getValue());
        } finally {
            synchronized (cellObjectMap) {
                processed++;
                cellObjectMap.notifyAll();
            }
        }
        return true;
    }

    @Override
    public Callable processStepAsynchronously() throws InterruptedException {
        final Entry<VoronoiLeafCell, List<LocalAbstractObject>> objectsToBeInserted = getNextProcessingItem();
        if (objectsToBeInserted == null)
            return null;
        return new Callable<AbstractOperation>() {
            @Override
            public AbstractOperation call() throws InterruptedException, AlgorithmMethodException {
                try {
                    processObjectsOnLeaf(objectsToBeInserted.getKey(), objectsToBeInserted.getValue());
                } finally {
                    synchronized (cellObjectMap) {
                        processed++;
                        cellObjectMap.notifyAll();
                    }
                }
                return operation;
            }
        };
    }
    
    /**
     * Insert the given objects into the given leaf cell. If some of the objects
     * could NOT be inserted (i.e. bucket is full) split the leaf cell.
     * @param objects to be inserted
     */
    protected void processObjectsOnLeaf(VoronoiLeafCell leaf, List<LocalAbstractObject> objects) throws AlgorithmMethodException {
        leaf.readLock();
        SplitConfiguration splitConfig = null;
        try {
            if (!leaf.isValid()) {
                for (LocalAbstractObject object : objects) {
                    addToCellObjectMap(object);
                }
                return;
            }
            splitConfig = leaf.insertObjects(objects);
        } finally {
            leaf.readUnLock();
        }
        if (splitConfig != null) { // if any splitConfig is returned, split the leaf
            // some of the objects weren't inserted into the leaf cell's bucket
            // so split the leaf cell, insert them into appropriate leaf cells.
            leaf.writeLock();
            try {
                leaf.split(splitConfig);
            } finally {
                leaf.writeUnLock();
            }
            
            // upper layers calling this processor need to handle proper insertion of objects after split by calling processStep() method
            for (LocalAbstractObject object : splitConfig.getToReinsert()) {
                addToCellObjectMap(object);
            }
        }
    }

    /**
     * A process step of this Processor means inserting all LocalAbstractObjects
     * that are mapped to a MIndexLeafCell into its Bucket
     * @return the number of LeafCells that were already processed
     */
    @Override
    public int getProcessedCount() {
        return processed;
    }

    /**
     * A process step of this Processor means inserting all LocalAbstractObjects
     * that are mapped to a MIndexLeafCell into its Bucket. After calling processNext()
     * method number of times this method returns hasProcessNext should still be called
     * as some more objects might need to be processed because of occurrence of a split.
     * @return the number of StructureLeafNodes that are to be processed
     */
    @Override
    public int getRemainingCount() {
        return cellObjectMap.size() - processed;
    }

    /**
     * If processing of the operation this NavigationProcessor holds is ended
     * prematurely - not all leaf nodes were processed - unlock the remaining
     * leaf nodes
     */
    @Override
    public void close() {
        for (VoronoiLeafCell leaf : cellObjectMap.keySet()) {
            leaf.readUnLock();
        }
    }

    @Override
    public AbstractOperation getOperation() {
        return this.operation;
    }

    /**
     * For the given INITIALIZED object (i.e. {@code MetricIndex.readMIndexPP(object)} != null) find the responsible leaf cell
     * add it to the global cell object map and if not already locked, acquire the read lock on the leaf cell
     * @param object to be added to the list for its responsible MIndexLeafCell
     * @throws AlgorithmMethodException 
     */
    private void addToCellObjectMap(LocalAbstractObject object) throws AlgorithmMethodException {
        VoronoiLeafCell leaf = mIndex.getVoronoiCellTree().getResponsibleLeaf(mIndex.readPPP(object), true);
        
        synchronized (cellObjectMap) {
            if (cellObjectMap.containsKey(leaf)) {
                cellObjectMap.get(leaf).add(object);
            } else {
                List<LocalAbstractObject> objects = new ArrayList<>();
                objects.add(object);
                cellObjectMap.put(leaf, objects);
            }
            cellObjectMap.notifyAll();
        }
    }

    /**
     * Computes MIndexKey for each of the objects and ands it to the list of its responsible leaf cell.
     * If allowed - initializing the objects (computing distances to pivots) is done in parallel.
     * @param objects to be initialized
     * @throws AlgorithmMethodException if anything goes wrong
     */
    private void initializeObjects(Collection<? extends LocalAbstractObject> objects) throws AlgorithmMethodException {
        for (LocalAbstractObject object : mIndex.initPP(objects)) {
            addToCellObjectMap(object);
        }
    }

    /**
     * Factory method for creating an instance of InsertOperationNavigationProcessor for given Bulk/InsertOperation
     * If allowed - initializing the objects (computing distances to pivots) is done in parallel. Acquires write lock on the leaf nodes.
     * @param operation - either instance of InsertOperation or BulkInsertOperation
     * @param mIndex
     * @return initialized instance of InsertOperationNavigationProcessor
     * @throws AlgorithmMethodException 
     */
    public static InsertOperationNavigationProcessor getInstance(AbstractOperation operation, MetricIndex mIndex) throws AlgorithmMethodException {
        if (operation instanceof InsertOperation) {
            return new InsertOperationNavigationProcessor((InsertOperation) operation, mIndex);
        } else if (operation instanceof BulkInsertOperation) {
            return new InsertOperationNavigationProcessor((BulkInsertOperation) operation, mIndex);
        } else {
            throw new UnsupportedOperationException("InsertOperationNavigationProcessor can be created only for InsertOperation or BulkInsertOperation");
        }
    }
}
